import logging
import requests

from lxml import etree as ET
from six.moves.urllib_parse import urlparse

from . import discovery
from .util import deserialize_apps, deserialize_channels

try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus


__version__ = '3.0.0'


COMMANDS = {
    # Standard Keys
    'home': 'Home',
    'reverse': 'Rev',
    'forward': 'Fwd',
    'play': 'Play',
    'select': 'Select',
    'left': 'Left',
    'right': 'Right',
    'down': 'Down',
    'up': 'Up',
    'back': 'Back',
    'replay': 'InstantReplay',
    'info': 'Info',
    'backspace': 'Backspace',
    'search': 'Search',
    'enter': 'Enter',
    'literal': 'Lit',

    # For devices that support "Find Remote"
    'find_remote': 'FindRemote',

    # For Roku TV
    'volume_down': 'VolumeDown',
    'volume_up': 'VolumeUp',
    'volume_mute': 'VolumeMute',

    # For Roku TV while on TV tuner channel
    'channel_up': 'ChannelUp',
    'channel_down': 'ChannelDown',

    # For Roku TV current input
    'input_tuner': 'InputTuner',
    'input_hdmi1': 'InputHDMI1',
    'input_hdmi2': 'InputHDMI2',
    'input_hdmi3': 'InputHDMI3',
    'input_hdmi4': 'InputHDMI4',
    'input_av1': 'InputAV1',

    # For devices that support being turned on/off
    'power': 'Power',
    'poweroff': 'PowerOff',
}

SENSORS = ('acceleration', 'magnetic', 'orientation', 'rotation')

TOUCH_OPS = ('up', 'down', 'press', 'move', 'cancel')


roku_logger = logging.getLogger('roku')


class RokuException(Exception):
    pass


class Application(object):

    def __init__(self, id, version, name, roku=None, is_screensaver=False):
        self.id = str(id)
        self.version = version
        self.name = name
        self.is_screensaver = is_screensaver
        self.roku = roku

    def __eq__(self, other):
        return isinstance(other, Application) and \
            (self.id, self.version) == (other.id, other.version)

    def __repr__(self):
        return ('<Application: [%s] %s v%s>' %
                (self.id, self.name, self.version))

    @property
    def icon(self):
        if self.roku:
            return self.roku.icon(self)

    def launch(self):
        if self.roku:
            self.roku.launch(self)

    def store(self):
        if self.roku:
            self.roku.store(self)


class Channel(object):

    def __init__(self, number, name, roku=None):
        self.number = str(number)
        self.name = name
        self.roku = roku

    def __eq__(self, other):
        return isinstance(other, Channel) and \
            (self.number, self.name) == (other.number, other.name)

    def __repr__(self):
        return ('<Channel: [%s] %s>' %
                (self.number, self.name))

    def launch(self):
        if self.roku:
            tv_app = Application(id= 'tvinput.dtv', version=None, name='TV', roku=self.roku)
            self.roku.launch(tv_app, {'ch': self.number})


class DeviceInfo(object):

    def __init__(self, model_name, model_num, software_version, serial_num, user_device_name, roku_type):
        self.model_name = model_name
        self.model_num = model_num
        self.software_version = software_version
        self.serial_num = serial_num
        self.user_device_name = user_device_name
        self.roku_type = roku_type

    def __repr__(self):
        return ('<DeviceInfo: %s-%s, SW v%s, Ser# %s (%s)>' %
                (self.model_name, self.model_num,
                 self.software_version, self.serial_num, self.roku_type))


class Roku(object):

    @classmethod
    def discover(self, *args, **kwargs):
        rokus = []
        for device in discovery.discover(*args, **kwargs):
            o = urlparse(device.location)
            rokus.append(Roku(o.hostname, o.port))
        return rokus

    def __init__(self, host, port=8060, timeout=10):
        self.host = host
        self.port = port
        self._conn = None
        self.timeout = timeout

    def __repr__(self):
        return "<Roku: %s:%s>" % (self.host, self.port)

    def __getattr__(self, name):

        if name not in COMMANDS and name not in SENSORS:
            raise AttributeError('%s is not a valid method' % name)

        def command(*args):
            if name in SENSORS:
                keys = ['%s.%s' % (name, axis) for axis in ('x', 'y', 'z')]
                params = dict(zip(keys, args))
                self.input(params)
            elif name == 'literal':
                for char in args[0]:
                    path = '/keypress/%s_%s' % (COMMANDS[name], quote_plus(char))
                    self._post(path)
            elif name == 'search':
                keys = ['title', 'season', 'launch', 'provider', 'type']
                params = dict(zip(keys, args))
                path = '/search/browse'
                self._post(path, params=params)
            else:
                path = '/keypress/%s' % COMMANDS[name]
                self._post(path)

        return command

    def __getitem__(self, key):
        key = str(key)
        app = self._app_for_name(key)
        if not app:
            app = self._app_for_id(key)
        return app

    def _app_for_name(self, name):
        for app in self.apps:
            if app.name == name:
                return app

    def _app_for_id(self, app_id):
        for app in self.apps:
            if app.id == app_id:
                return app

    def _connect(self):
        if self._conn is None:
            self._conn = requests.Session()

    def _get(self, path, *args, **kwargs):
        return self._call('GET', path, *args, **kwargs)

    def _post(self, path, *args, **kwargs):
        return self._call('POST', path, *args, **kwargs)

    def _call(self, method, path, *args, **kwargs):

        self._connect()

        roku_logger.debug(path)

        url = 'http://%s:%s%s' % (self.host, self.port, path)

        if method not in ('GET', 'POST'):
            raise ValueError('only GET and POST HTTP methods are supported')

        func = getattr(self._conn, method.lower())
        resp = func(url, timeout=self.timeout, *args, **kwargs)

        if resp.status_code not in [200, 204]:
            raise RokuException(resp.content)

        return resp.content

    @property
    def apps(self):
        resp = self._get('/query/apps')
        applications = deserialize_apps(resp)
        for a in applications:
            a.roku = self
        return applications

    @property
    def active_app(self):
        resp = self._get('/query/active-app')
        active_app = deserialize_apps(resp)
        if len(active_app):
            return active_app[0]
        else:
            return None

    @property
    def tv_channels(self):
        resp = self._get('/query/tv-channels')
        channels = deserialize_channels(resp)
        for c in channels:
            c.roku = self
        return channels

    @property
    def device_info(self):
        resp = self._get('/query/device-info')
        root = ET.fromstring(resp)

        roku_type = "Box"
        if root.find('is-tv') != None and root.find('is-tv').text == "true":
            roku_type = "TV"
        elif root.find('is-stick') != None and \
             root.find('is-stick').text == "true":
            roku_type = "Stick"
        dinfo = DeviceInfo(
            model_name=root.find('model-name').text,
            model_num=root.find('model-number').text,
            software_version=''.join([
                root.find('software-version').text,
                '.',
                root.find('software-build').text
            ]),
            serial_num=root.find('serial-number').text,
            user_device_name=root.find('user-device-name').text,
            roku_type=roku_type
        )
        return dinfo

    @property
    def commands(self):
        return sorted(COMMANDS.keys())

    @property
    def power_state(self):
        resp = self._get('/query/device-info')
        root = ET.fromstring(resp)
        if root.find('power-mode').text:
            if root.find('power-mode').text == 'PowerOn':
                return "On"
            else:
                return "Off"
        return "Unknown"

    def icon(self, app):
        return self._get('/query/icon/%s' % app.id)

    def launch(self, app, params={}):
        if app.roku and app.roku != self:
            raise RokuException('this app belongs to another Roku')
        params['contentID'] = app.id
        return self._post('/launch/%s' % app.id, params=params)

    def store(self, app):
        return self._post('/launch/11', params={'contentID': app.id})

    def input(self, params):
        return self._post('/input', params=params)

    def touch(self, x, y, op='down'):

        if op not in TOUCH_OPS:
            raise RokuException('%s is not a valid touch operation' % op)

        params = {
            'touch.0.x': x,
            'touch.0.y': y,
            'touch.0.op': op,
        }

        self.input(params)

    @property
    def current_app(self):
        resp = self._get('/query/active-app')
        root = ET.fromstring(resp)
        is_screensaver = True

        app_node = root.find('screensaver')
        if app_node is None:
            app_node = root.find('app')
            is_screensaver = False

        if app_node is None:
            return None

        return Application(
            id=app_node.get('id'),
            version=app_node.get('version'),
            name=app_node.text,
            is_screensaver=is_screensaver,
            roku=self,
        )



from flask import Flask, request

app = Flask(__name__)


@app.route('/keypress/<key>')
def keypress(key):
    pass


@app.route('/launch/<code>')
def launch(code):
    app_id = request.args.get('contentID', None)
    pass


@app.route('/query/apps')
def list_apps():
    pass


@app.route('/query/active-app')
def active_app():
    pass


@app.route('/query/icon/<app_id>')
def app_icon(app_id):
    pass


if __name__ == '__main__':
    app.run(port=8060, debug=True)



from roku import Application

DEFAULT_APPS = [
    Application(1, '1.0', 'Hulu Plus'),
    Application(2, '2.0', 'TWiT'),
    Application(3, '3.0', 'Whisky Media'),
    Application(4, '4.0', 'Netflix'),
]


class Emulator(object):

    def __init__(self, apps=None):
        self._apps = apps or DEFAULT_APPS

    def __call__(self, command, *args, **kwargs):
        pass

    def add_app(self, app):
        pass

    def get_icon(self, app_id):
        pass

    def launch_app(self, app_id):
        pass

    def list_apps(self):
        pass



from contextlib import closing

from lxml import etree as ET
from six import BytesIO


def deserialize_apps(doc, roku=None):

    from .core import Application

    applications = []
    root = ET.fromstring(doc)
    for elem in root:
        app = Application(
            id=elem.get('id'), version=elem.get('version'), name=elem.text)
        applications.append(app)
    return applications


def serialize_apps(apps):

    root = ET.Element('apps')

    for app in apps:
        attrs = {'id': app.id, 'version': app.version}
        elem = ET.SubElement(root, 'app', attrs)
        elem.text = app.name

    with closing(BytesIO()) as bffr:
        tree = ET.ElementTree(root)
        tree.write(bffr, xml_declaration=True, encoding="utf-8")
        content = bffr.getvalue()

    return content

def deserialize_channels(doc, roku=None):

    from .core import Channel

    channels = []
    root = ET.fromstring(doc)

    for elem in root:
        channel = Channel(
            number=elem.find('number').text,
            name=elem.find('name').text,
            roku=roku,
        )
        channels.append(channel)
    return channels



# -*- coding: utf-8 -*-
import os

import pytest

try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from roku.core import Application, Roku, COMMANDS
from roku.util import serialize_apps


TESTS_PATH = os.path.abspath(os.path.dirname(__file__))


class Fauxku(Roku):

    def __init__(self, *args, **kwargs):
        super(Fauxku, self).__init__(*args, **kwargs)
        self._calls = []

    def _call(self, method, path, *args, **kwargs):
        self._calls.append((method, path, args, kwargs))
        return ''

    def calls(self):
        return self._calls

    def last_call(self):
        return self._calls[-1]


@pytest.fixture
def roku():
    return Fauxku('0.0.0.0')


@pytest.fixture
def apps(roku):
    faux_apps = [
        Application('11', '1.0.1', 'Fauxku Channel Store', roku),
        Application('22', '2.0.2', 'Faux Netflix', roku),
        Application('33', '3.0.3', 'Faux YouTube', roku),
        Application('44HL', '4.0.4', 'Faux Hulu', roku),
    ]
    return faux_apps


def test_apps(mocker, roku):

    faux_apps = (Application('0x', '1.2.3', 'Fauxku Channel Store'),)

    mocked_get = mocker.patch.object(Roku, '_get')
    mocked_get.return_value = serialize_apps(faux_apps)

    apps = roku.apps

    assert len(apps) == 1
    assert apps[0].id == '0x'
    assert apps[0].version == '1.2.3'
    assert apps[0].name == 'Fauxku Channel Store'


def test_app_selector(mocker, apps):

    mocked_get = mocker.patch.object(Roku, '_get')
    mocked_get.return_value = serialize_apps(apps)

    for app in apps:
        roku = app.roku
        assert roku[app.id] == app
        assert roku[app.name] == app


def test_device_info(mocker, roku):

    xml_path = os.path.join(TESTS_PATH, 'responses', 'device-info.xml')
    with open(xml_path) as infile:
        content = infile.read()

    mocked_get = mocker.patch.object(Roku, '_get')
    mocked_get.return_value = content.encode('utf-8')

    d = roku.device_info

    assert d.model_name == 'Roku 3'
    assert d.model_num == '4200X'
    assert d.software_version == '7.00.09044'
    assert d.serial_num == '111111111111'
    assert d.roku_type == 'Stick'


def test_commands(roku):

    for cmd in roku.commands:

        if cmd in ['literal', 'search']:
            # there is a separate test for the literal and search command
            continue

        getattr(roku, cmd)()
        call = roku.last_call()

        assert call == ('POST', '/keypress/%s' % COMMANDS[cmd], (), {})


def test_search(roku):

    text = 'Stargate'
    roku.search(text)

    call = roku.last_call()

    assert call == ('POST', '/search/browse', (), {'params': {'title': text}})


def test_literal(roku):

    text = 'Stargate'
    roku.literal(text)

    for i, call in enumerate(roku.calls()):
        assert call == ('POST', '/keypress/Lit_%s' % quote_plus(text[i]), (), {})


def test_literal_fancy(roku):

    text = r"""~!@#$%^&*()_+`-=[]{};':",./<>?\|€£"""
    roku.literal(text)

    for i, call in enumerate(roku.calls()):
        assert call == ('POST', '/keypress/Lit_%s' % quote_plus(text[i]), (), {})


def test_store(apps):

    for app in apps:

        roku = app.roku
        roku.store(app)
        call = roku.last_call()

        params = {'params': {'contentID': app.id}}
        assert call == ('POST', '/launch/11', (), params)


def test_launch(apps):

    for app in apps:

        roku = app.roku
        roku.launch(app)
        call = roku.last_call()

        params = {'params': {'contentID': app.id}}
        assert call == ('POST', '/launch/%s' % app.id, (), params)



"""
Code adapted from Dan Krause.
https://gist.github.com/dankrause/6000248
http://github.com/dankrause
"""
import socket
import six
from six.moves import http_client

ST_DIAL = 'urn:dial-multiscreen-org:service:dial:1'
ST_ECP = 'roku:ecp'


class _FakeSocket(six.BytesIO):
    def makefile(self, *args, **kw):
        return self


class SSDPResponse(object):

    def __init__(self, response):
        self.location = response.getheader('location')
        self.usn = response.getheader('usn')
        self.st = response.getheader('st')
        self.cache = response.getheader('cache-control').split('=')[1]

    def __repr__(self):
        return '<SSDPResponse({location}, {st}, {usn})'.format(**self.__dict__)


def discover(timeout=2, retries=1, st=ST_ECP):

    group = ('239.255.255.250', 1900)

    message = '\r\n'.join([
        'M-SEARCH * HTTP/1.1',
        'HOST: {0}:{1}'.format(*group),
        'MAN: "ssdp:discover"',
        'ST: {st}', 'MX: 3', '', ''])

    socket.setdefaulttimeout(timeout)

    responses = {}

    for _ in range(retries):
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)

        m = message.format(st=st)
        if six.PY2:
            sock.sendto(m, group)
        elif six.PY3:
            sock.sendto(m.encode(), group)

        while 1:
            try:
                rhttp = http_client.HTTPResponse(_FakeSocket(sock.recv(1024)))
                rhttp.begin()
                if rhttp.status == 200:
                    rssdp = SSDPResponse(rhttp)
                    responses[rssdp.location] = rssdp
            except socket.timeout:
                break

    return responses.values()



from setuptools import setup, find_packages
import os

f = open(os.path.join(os.path.dirname(__file__), 'README.rst'))
readme = f.read()
f.close()

setup(
    name='roku',
    version='3.0',
    description='Client for the Roku media player',
    long_description=readme,
    author='Jeremy Carbaugh',
    author_email='jcarbaugh@gmail.com',
    url='https://github.com/jcarbaugh/python-roku',
    packages=find_packages(),
    install_requires=[
        'lxml>=3.6,<3.7',
        'requests>=2.10,<2.11',
        'six'
    ],
    license='BSD License',
    platforms=["any"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)
